#include <stdio.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>

// ===================================================================
// Universal Indexer for .ts files streamed with Live555 RTSP Server
// Sebastien RAILLARD - COEXSI SARL 2010 - www.coexsi.fr
// ===================================================================

// Main entry point
int main(int argc, char const** argv) 
{
	// Test parameter count
	if (argc!=2)
	{
		printf("ERROR: One input .ts file must be specified as parameter!\n");
		return(1);
	}
	char const* FileName = argv[1];

	// Test if file exists
	struct stat info;
	if (stat(FileName,&info)!=0)
	{
		printf("ERROR: Input file [%s] doesn't exist!\n",FileName);
		return(1);
	}
	
	// Check whether the input file name ends with ".ts":
	int len = strlen(FileName);
	if (len<4 || strcmp(&FileName[len-3], ".ts") != 0)
	{
		printf("ERROR: Input file [%s] isn't a .ts file!\n",FileName);
		return(1);
	}
	
	// Create index file name .ts => .tsx
	char* IndexName;
	asprintf(&IndexName, "%sx", FileName);
	
	// Infos
	printf("Starting indexing of [%s]\n",FileName);
	printf("Index file created will be [%s]\n",IndexName);
	
	// Some variables declaration
	// long long unsigned int = uint64 
	// unsigned int = uint32
	size_t NbRead = 0;
	long long unsigned int Packets = 0;
	unsigned int State = 0; // 0=Search PMT 1=Search PCR 2=Search PES
	unsigned int PCR_PID = 0;
	double First_PCR = 0;
	double Last_PCR = 0;
	double Index_PCR = 0;
	long long unsigned int PES_Found = 0;
	long long unsigned int PES_filtered = 0;
	time_t StartTime = time(NULL);
	unsigned char Buffer[188];

	// Opening files
	FILE *sr;
	sr = fopen(FileName,"rb");
	if (sr==NULL)
	{
		printf("ERROR: Cannot open source file [%s]!\n",FileName);
		return(1);
	}
	FILE *sw;
	sw = fopen(IndexName,"wb");
	if (sw==NULL)
	{
		printf("ERROR: Cannot open index file [%s]!\n",IndexName);
		fclose(sr);
		return(1);
	}
	
	// Loop
	printf("Searching the first PSI-PMT packet...\n");
	do
	{
		// Read the next TS packet and check its size
		NbRead=fread(Buffer,1,sizeof(Buffer),sr);
		
		if (NbRead==188)
		{
			// Starting TS packet analysis
			Packets++;
			
			// General TS values
			unsigned char sync_byte = Buffer[0];
			unsigned char transport_error_indicator = (Buffer[1] & 0x80) >> 7;
			unsigned char payload_unit_start_indicator = (Buffer[1] & 0x40) >> 6;
			unsigned int PID = (((unsigned int)Buffer[1] & 0x1F) << 8) + (unsigned int)Buffer[2];
			unsigned char transport_scrambling_control = (Buffer[3] & 0xC0) >> 6;
			unsigned char adaptation_field_control = (Buffer[3] & 0x30) >> 4;

			// Check if the sync byte is present
			if (sync_byte!=0x47) 
			{
				printf("ERROR: Missing the TS sync byte!\n");
				NbRead = 0;
			}
			
			// Searching PMT
			if (State == 0)
			{
				// TS values checking
				if (transport_error_indicator == 0 && payload_unit_start_indicator == 1 && PID != 0x1FFF && transport_scrambling_control == 0 && adaptation_field_control == 1)
				{
					// Checking if it's not a PES - No packet_start_code_prefix
					if (!(Buffer[4] == 0 && Buffer[5] == 0 && Buffer[6] == 1))
					{
						// Checking if there is PSI table signature with table_id=2 for PMT and we assume the pointer_field=0 (maybe not always true...)
						if (Buffer[4] == 0 && Buffer[5] == 2 && (Buffer[6] & 0xC0) == 0x80)
						{
							// Ok, we have the PMT, we need will get the PCR PID and assume that it is the Video PID
							printf(" * PMT PID is %d (0x%X) - %llu Packets skipped since beginning\n",PID,PID,(Packets - 1));
							PCR_PID = (((unsigned int)Buffer[13] & 0x1F) << 8) + (unsigned int)Buffer[14];
							printf(" * PCR PID is %d (0x%X) - We assume it is also the video PID\n",PCR_PID,PCR_PID);
							// Next, we search the first PCR
							State = 1;
							printf("Searching the first PCR...\n");
						}
					}
				}
			}
			
			// Searching the PCR
			if (State == 1 || State == 2)
			{
				// Checking some TS values - Looking for PCR PID with adpation field
				if (PID == PCR_PID && transport_error_indicator == 0 && (adaptation_field_control == 2 || adaptation_field_control == 3))
				{
					// Checking if adpation field length is not nul and if PCR flag is set
					if (Buffer[4] > 0 && (Buffer[5] & 0x10) == 0x10)
					{
						// Getting the raw PCR values
						unsigned long long int PCR_Base = ((unsigned long long int)Buffer[6] << 25) + ((unsigned long long int)Buffer[7] << 17) + ((unsigned long long int)Buffer[8] << 9) + ((unsigned long long int)Buffer[9] << 1) + (((unsigned long long int)Buffer[10] & 0x80) >> 7);
						unsigned long long int PCR_Ext = (((unsigned long long int)Buffer[10] & 0x1) << 8) + (unsigned long long int)Buffer[11];
						// Transforming to double (float isn't big enough!)
						Last_PCR = ((double)PCR_Base / 90000.0f) + ((double)PCR_Ext / 27000000.0f);
						// Check if we were searching the first PCR
						if (State == 1)
						{
							// We start indexing
							State = 2;
							First_PCR = Last_PCR;
							printf(" * First PCR value is %.3fs - %llu Packets skipped since beginning\n",First_PCR,(Packets - 1));
							printf("Searching PES sections start for indexing...\n");
						}
					}
				}
			}
			
			// Indexing 
			if (State == 2)
			{
				// Checking some TS values - Looking for PCR PID with start section
				if (PID == PCR_PID && transport_error_indicator == 0 && payload_unit_start_indicator == 1)
				{
					// We certainly have a PES section start here
					PES_Found++;
					// Check if we have a new PCR
					if (Index_PCR != Last_PCR)
					{
						// We keep the value for the next check
						Index_PCR = Last_PCR;
						// We create an index
						unsigned char Index[11];
						Index[0] = 0x81; // RECORD_VSH / First record of frame
						Index[1] = 0; // No start offset information
						Index[2] = 0; // No payload size information
						// PCR transformation - Starting at 0
						unsigned long long int PCR_Int = (unsigned long long int)(Last_PCR - First_PCR); // Truncate float to integer
						Index[3] = (unsigned char)(PCR_Int & 0xFF);
						Index[4] = (unsigned char)((PCR_Int & 0xFF00) >> 8);
						Index[5] = (unsigned char)((PCR_Int & 0xFF0000) >> 16);
						unsigned char PCR_Frac = (unsigned char)(((Last_PCR - First_PCR) - (double)(PCR_Int)) * 255.0f);
						Index[6] = PCR_Frac;
						// TS packet number - Starting at 1
						Index[7] = (unsigned char)(Packets & 0xFF);
						Index[8] = (unsigned char)((Packets & 0xFF00) >> 8);
						Index[9] = (unsigned char)((Packets & 0xFF0000) >> 16);
						Index[10] = (unsigned char)((Packets & 0xFF000000) >> 24);
						// Writing the index
						fwrite(Index,1,sizeof(Index),sw);
					}
					else
					{
						// We filter this PES (no PCR update)
						PES_filtered++;
					}
				}
			}

		}

		if (NbRead>0 && NbRead<188)
		{
			// Check if we have a uncommon read size
			printf("WARNING: The file doesn't contain an exact number of TS packets!\n");
		}
	}
	while(NbRead==188);
	
	// Closing files
	fclose(sr);
	fclose(sw);
	
	// Show stats
	printf(" * %llu TS packets were read\n",Packets);
	printf(" * %llu PES sections were found\n",PES_Found);
	printf(" * %llu PES sections were filtered\n",PES_filtered);
	if (First_PCR > 0 && Last_PCR > 0)
	{
		printf(" * Last PCR value is %.3fs\n",Last_PCR);
		printf(" * Movie length estimation is %.3fs\n",(Last_PCR - First_PCR));
	}
	printf(" * Indexing running time was %.0fs\n",difftime(time(NULL),StartTime));

	// End
	printf("End of indexing.\n");
	return(0);
}
